/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TDecSbac.h
    \brief    SBAC decoder class (header)
*/

#ifndef __TDECSBAC__
#define __TDECSBAC__


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "TDecEntropy.h"
#include "TDecBinCoder.h"
#include "TLibCommon/ContextTables.h"
#include "TLibCommon/ContextModel.h"
#include "TLibCommon/ContextModel3DBuffer.h"

//! \ingroup TLibDecoder
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

class SEImessages;

/// SBAC decoder class
class TDecSbac : public TDecEntropyIf
{
public:
  TDecSbac();
  virtual ~TDecSbac();
  
  Void  init                      ( TDecBinIf* p )    { m_pcTDecBinIf = p; }
  Void  uninit                    (              )    { m_pcTDecBinIf = 0; }
  
  Void load                          ( TDecSbac* pScr );
  Void loadContexts                  ( TDecSbac* pScr );
  Void xCopyFrom           ( TDecSbac* pSrc );
  Void xCopyContextsFrom       ( TDecSbac* pSrc );
  Void decodeFlush();

  Void  resetEntropy (TComSlice* pSlice );
  Void  setBitstream              ( TComInputBitstream* p  ) { m_pcBitstream = p; m_pcTDecBinIf->init( p ); }
#if AHG5_COUNT_XXX_BITS_BINS
  UInt  getNumberOfLeftBits()    { return m_pcBitstream->getNumBitsLeft() + m_pcTDecBinIf->getNumberOfHoldingBits(); }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_ABS_BITS_BINS
  Void countSAOABSBits                ( UInt ui ) { m_countedSAOABSBits += ui; }
  Void countSAOABSContextBins         ( UInt ui ) { m_countedSAOABSContextBins += ui; }
  Void countSAOABSBypassBins          ( UInt ui ) { m_countedSAOABSBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_TYPE_BITS_BINS
  Void countSAOTypeBits                ( UInt ui ) { m_countedSAOTypeBits += ui; }
  Void countSAOTypeContextBins         ( UInt ui ) { m_countedSAOTypeContextBins += ui; }
  Void countSAOTypeBypassBins          ( UInt ui ) { m_countedSAOTypeBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_ENABLE_FLAG_BITS_BINS
  Void countSAOEFBits                ( UInt ui ) { m_countedSAOEFBits += ui; }
  Void countSAOEFContextBins         ( UInt ui ) { m_countedSAOEFContextBins += ui; }
  Void countSAOEFBypassBins          ( UInt ui ) { m_countedSAOEFBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_MERGE_LEFT_BITS_BINS
  Void countSAOMLBits                ( UInt ui ) { m_countedSAOMLBits += ui; }
  Void countSAOMLContextBins         ( UInt ui ) { m_countedSAOMLContextBins += ui; }
  Void countSAOMLBypassBins          ( UInt ui ) { m_countedSAOMLBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_MERGE_UP_BITS_BINS
  Void countSAOMUBits                ( UInt ui ) { m_countedSAOMUBits += ui; }
  Void countSAOMUContextBins         ( UInt ui ) { m_countedSAOMUContextBins += ui; }
  Void countSAOMUBypassBins          ( UInt ui ) { m_countedSAOMUBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_LEFT_BAND_BITS_BINS
  Void countSAOLBBits                ( UInt ui ) { m_countedSAOLBBits += ui; }
  Void countSAOLBContextBins         ( UInt ui ) { m_countedSAOLBContextBins += ui; }
  Void countSAOLBBypassBins          ( UInt ui ) { m_countedSAOLBBypassBins += ui; }
#endif
#if AHG5_COUNT_XXX_BITS_BINS && AHG5_COUNT_SAO_SIGN_BITS_BINS
  Void countSAOSIGNBits                ( UInt ui ) { m_countedSAOSIGNBits += ui; }
  Void countSAOSIGNContextBins         ( UInt ui ) { m_countedSAOSIGNContextBins += ui; }
  Void countSAOSIGNBypassBins          ( UInt ui ) { m_countedSAOSIGNBypassBins += ui; }
#endif

  Void  parseSPS                  ( TComSPS* pcSPS         ) {}
  Void  parsePPS                  ( TComPPS* pcPPS, ParameterSetManagerDecoder *parameterSet         ) {}
  Void  parseAPS                  ( TComAPS* pAPS          ) {}
  void parseSEI(SEImessages&) {}

#if AHG6_ALF_OPTION2
  Void  parseSliceHeader          ( TComSlice*& rpcSlice, ParameterSetManagerDecoder *parameterSetManager) {}
#else
  Void  parseSliceHeader          ( TComSlice*& rpcSlice, ParameterSetManagerDecoder *parameterSetManager, AlfCUCtrlInfo &alfCUCtrl, AlfParamSet& alfParamSet) {}
#endif
  Void  parseTerminatingBit       ( UInt& ruiBit );
  Void  parseMVPIdx               ( Int& riMVPIdx          );
#if SAO_TRUNCATED_U
  Void  parseSaoMaxUvlc           ( UInt& val, UInt maxSymbol );
#endif
  Void  parseSaoUvlc              ( UInt& ruiVal           );
#if !(SAO_OFFSET_MAG_SIGN_SPLIT && SAO_RDO_FIX)
  Void  parseSaoSvlc              ( Int&  riVal            );
#endif
  Void  parseSaoMergeLeft         ( UInt&  ruiVal, UInt uiCompIdx   );
  Void  parseSaoMergeUp           ( UInt&  ruiVal  );
  Void  parseSaoTypeIdx           ( UInt&  ruiVal  );
  Void  parseSaoUflc              ( UInt& ruiVal           );
#if SAO_NO_MERGE_CROSS_SLICE_TILE
  Void  parseSaoOneLcuInterleaving(Int rx, Int ry, SAOParam* pSaoParam, TComDataCU* pcCU, Int iCUAddrInSlice, Int iCUAddrUpInSlice, Int allowMergeLeft, Int allowMergeUp);
#else
  Void  parseSaoOneLcuInterleaving(Int rx, Int ry, SAOParam* pSaoParam, TComDataCU* pcCU, Int iCUAddrInSlice, Int iCUAddrUpInSlice, Bool bLFCrossSliceBoundaryFlag);
#endif
  Void  parseSaoOffset            (SaoLcuParam* psSaoLcuParam);
private:
  Void  xReadUnarySymbol    ( UInt& ruiSymbol, ContextModel* pcSCModel, Int iOffset );
  Void  xReadUnaryMaxSymbol ( UInt& ruiSymbol, ContextModel* pcSCModel, Int iOffset, UInt uiMaxSymbol );
  Void  xReadEpExGolomb     ( UInt& ruiSymbol, UInt uiCount );
#if COEF_REMAIN_BINARNIZATION
  Void  xReadCoefRemainExGolomb ( UInt &rSymbol, UInt &rParam );
#else
  Void  xReadGoRiceExGolomb ( UInt &ruiSymbol, UInt &ruiGoRiceParam );
#endif
private:
  TComInputBitstream* m_pcBitstream;
  TDecBinIf*        m_pcTDecBinIf;
  
  Int           m_iSliceGranularity; //!< slice granularity

public:
#if AHG6_ALF_OPTION2
  Void parseAlfCtrlFlag   (Int compIdx, UInt& code);
#else
  Void parseAlfCtrlFlag   ( UInt &ruiAlfCtrlFlag );
#endif
  /// set slice granularity
  Void setSliceGranularity(Int iSliceGranularity)  {m_iSliceGranularity = iSliceGranularity;}

  /// get slice granularity
  Int  getSliceGranularity()                       {return m_iSliceGranularity;             }

  Void parseSkipFlag      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  Void parseSplitFlag     ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  Void parseMergeFlag     ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth, UInt uiPUIdx );
  Void parseMergeIndex    ( TComDataCU* pcCU, UInt& ruiMergeIndex, UInt uiAbsPartIdx, UInt uiDepth );
  Void parsePartSize      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  Void parsePredMode      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  
  Void parseIntraDirLumaAng( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  
  Void parseIntraDirChroma( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  
  Void parseInterDir      ( TComDataCU* pcCU, UInt& ruiInterDir, UInt uiAbsPartIdx, UInt uiDepth );
  Void parseRefFrmIdx     ( TComDataCU* pcCU, Int& riRefFrmIdx, UInt uiAbsPartIdx, UInt uiDepth, RefPicList eRefList );
  Void parseMvd           ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiPartIdx, UInt uiDepth, RefPicList eRefList );
  
  Void parseTransformSubdivFlag( UInt& ruiSubdivFlag, UInt uiLog2TransformBlockSize );
  Void parseQtCbf         ( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth, UInt uiDepth );
  Void parseQtRootCbf     ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth, UInt& uiQtRootCbf );
  
  Void parseDeltaQP       ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  
  Void parseIPCMInfo      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth);

  Void parseLastSignificantXY( UInt& uiPosLastX, UInt& uiPosLastY, Int width, Int height, TextType eTType, UInt uiScanIdx );
  Void parseCoeffNxN      ( TComDataCU* pcCU, TCoeff* pcCoef, UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType );
#if INTRA_TRANSFORMSKIP
  Void parseTransformSkipFlags ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt width, UInt height, UInt uiDepth, TextType eTType);
#endif

  Void readTileMarker   ( UInt& uiTileIdx, UInt uiBitsUsed );
  Void updateContextTables( SliceType eSliceType, Int iQp );

  Void  parseScalingList ( TComScalingList* scalingList ) {}

private:
  UInt m_uiLastDQpNonZero;
  UInt m_uiLastQp;
  
  ContextModel         m_contextModels[MAX_NUM_CTX_MOD];
  Int                  m_numContextModels;
  ContextModel3DBuffer m_cCUSplitFlagSCModel;
  ContextModel3DBuffer m_cCUSkipFlagSCModel;
  ContextModel3DBuffer m_cCUMergeFlagExtSCModel;
  ContextModel3DBuffer m_cCUMergeIdxExtSCModel;
  ContextModel3DBuffer m_cCUPartSizeSCModel;
  ContextModel3DBuffer m_cCUPredModeSCModel;
  ContextModel3DBuffer m_cCUAlfCtrlFlagSCModel;
  ContextModel3DBuffer m_cCUIntraPredSCModel;
  ContextModel3DBuffer m_cCUChromaPredSCModel;
  ContextModel3DBuffer m_cCUDeltaQpSCModel;
  ContextModel3DBuffer m_cCUInterDirSCModel;
  ContextModel3DBuffer m_cCURefPicSCModel;
  ContextModel3DBuffer m_cCUMvdSCModel;
  ContextModel3DBuffer m_cCUQtCbfSCModel;
  ContextModel3DBuffer m_cCUTransSubdivFlagSCModel;
  ContextModel3DBuffer m_cCUQtRootCbfSCModel;
  
  ContextModel3DBuffer m_cCUSigCoeffGroupSCModel;
  ContextModel3DBuffer m_cCUSigSCModel;
  ContextModel3DBuffer m_cCuCtxLastX;
  ContextModel3DBuffer m_cCuCtxLastY;
  ContextModel3DBuffer m_cCUOneSCModel;
  ContextModel3DBuffer m_cCUAbsSCModel;
  
  ContextModel3DBuffer m_cMVPIdxSCModel;
  
  ContextModel3DBuffer m_cALFFlagSCModel;
  ContextModel3DBuffer m_cALFUvlcSCModel;
  ContextModel3DBuffer m_cALFSvlcSCModel;
  ContextModel3DBuffer m_cCUAMPSCModel;
  ContextModel3DBuffer m_cSaoFlagSCModel;
  ContextModel3DBuffer m_cSaoUvlcSCModel;
#if !(SAO_OFFSET_MAG_SIGN_SPLIT && SAO_RDO_FIX)
  ContextModel3DBuffer m_cSaoSvlcSCModel;
#endif
  ContextModel3DBuffer m_cSaoMergeLeftSCModel;
  ContextModel3DBuffer m_cSaoMergeUpSCModel;
  ContextModel3DBuffer m_cSaoTypeIdxSCModel;
#if INTRA_TRANSFORMSKIP
  ContextModel3DBuffer m_cTransformSkipSCModel;
#endif
};

//! \}

#endif // !defined(AFX_TDECSBAC_H__CFCAAA19_8110_47F4_9A16_810C4B5499D5__INCLUDED_)
